@@ -184,7 +184,8 @@ pub trait Entity<M>: Send {
     }
 
     /// Optional callback for running actions when the entity's owning [Facet] terminates
-    /// cleanly. Will not be called in case of abnormal shutdown (crash) of an actor.
+    /// cleanly. Will not be called in case of abnormal shutdown (crash) of an actor. Callbacks
+    /// run in the context of the owning facet's *parent* facet.
     ///
     /// Programs register an entity's stop hook with [Activation::on_stop_notify].
     ///
@@ -658,6 +659,24 @@ impl<'activation> Activation<'activation> {
         self.state.get_facet(self.facet.facet_id)
     }
 
+    /// Retrieves the chain of facet IDs, in order, from the currently-active [`Facet`] up to
+    /// and including the root facet of the active actor. Useful for debugging.
+    pub fn facet_ids(&mut self) -> Vec<FacetId> {
+        let mut ids = Vec::new();
+        let mut id = self.facet.facet_id;
+        loop {
+            ids.push(id);
+            match self.state.get_facet(id) {
+                None => break,
+                Some(f) => match f.parent_facet_id {
+                    None => break,
+                    Some(p) => id = p,
+                }
+            }
+        }
+        ids
+    }
+
     /// Core API: assert `a` at recipient `r`.
     ///
     /// Returns the [`Handle`] for the new assertion.
@@ -799,7 +818,8 @@ impl<'activation> Activation<'activation> {
     }
 
     /// Registers the entity `r` in the list of stop actions for the active facet. If the facet
-    /// terminates cleanly, `r`'s [`stop`][Entity::stop] will be called.
+    /// terminates cleanly, `r`'s [`stop`][Entity::stop] will be called in the context of the
+    /// facet's parent.
     ///
     /// **Note.** If the actor crashes, stop actions will *not* be called.
     ///
@@ -814,10 +834,18 @@ impl<'activation> Activation<'activation> {
     }
 
     /// Registers `action` in the list of stop actions for the active facet. If the facet
-    /// terminates cleanly, `action` will be called. See also notes against
-    /// [`on_stop_notify`][Activation::on_stop_notify].
+    /// terminates cleanly, `action` will be called in the context of the facet's parent. See
+    /// also notes against [`on_stop_notify`][Activation::on_stop_notify].
     pub fn on_stop<F: 'static + Send + FnOnce(&mut Activation) -> ActorResult>(&mut self, action: F) {
-        if let Some(f) = self.active_facet() {
+        self.on_facet_stop(self.facet.facet_id, action)
+    }
+
+    fn on_facet_stop<F: 'static + Send + FnOnce(&mut Activation) -> ActorResult>(
+        &mut self,
+        facet_id: FacetId,
+        action: F,
+    ) {
+        if let Some(f) = self.state.get_facet(facet_id) {
             f.stop_actions.push(Box::new(action));
         }
     }
@@ -1059,29 +1087,18 @@ impl<'activation> Activation<'activation> {
         }
     }
 
-    /// Arranges for the [`Facet`] named by `facet_id` to be stopped cleanly when `self`
-    /// commits.
-    ///
-    /// Then,
-    ///  - if `continuation` is supplied, and
-    ///  - the facet to be stopped hasn't been stopped at the time of the `stop_facet_and_continue` call, and
-    ///  - none of the shutdown handlers yields an error, and
-    ///  - the facet's parent facet is alive,
-    /// executes `continuation` (immediately) in the *parent* facet's context.
-    ///
+    /// If `continuation` is supplied, adds it as a stop action for the [`Facet`] named by
+    /// `facet_id`. Then, cleanly stops the facet immediately, without waiting for `self` to
+    /// commit.
     pub fn stop_facet_and_continue<F: 'static + Send + FnOnce(&mut Activation) -> ActorResult>(
         &mut self,
         facet_id: FacetId,
         continuation: Option<F>,
     ) -> ActorResult {
-        let maybe_parent_id = self.active_facet().and_then(|f| f.parent_facet_id);
-        self.enqueue_for_myself_at_commit(Box::new(move |t| t._terminate_facet(facet_id, true)));
         if let Some(k) = continuation {
-            if let Some(parent_id) = maybe_parent_id {
-                self.with_facet(true, parent_id, k)?;
-            }
+            self.on_facet_stop(facet_id, k);
         }
-        Ok(())
+        self._terminate_facet(facet_id, true)
     }
 
     /// Arranges for the [`Facet`] named by `facet_id` to be stopped cleanly when `self`
@@ -1131,10 +1148,13 @@ impl<'activation> Activation<'activation> {
                     }
                 }
                 if alive {
-                    for action in std::mem::take(&mut f.stop_actions) {
-                        action(t)?;
-                    }
                     let parent_facet_id = f.parent_facet_id;
+                    t.with_facet(false, parent_facet_id.unwrap_or(facet_id), |t| {
+                        for action in std::mem::take(&mut f.stop_actions) {
+                            action(t)?;
+                        }
+                        Ok(())
+                    })?;
                     f.retract_outbound(t);
                     // ^ we need retraction to happen right here so that child-facet
                     // cleanup-actions are performed before parent-facet cleanup-actions.
@@ -1685,9 +1705,9 @@ impl RunningActor {
         }
     }
 
-    /// Registers the entity `r` in the list of exit hooks for this
-    /// actor. When the actor terminates, `r`'s
-    /// [`exit_hook`][Entity::exit_hook] will be called.
+    /// Registers the entity `r` in the list of exit hooks for this actor. When the actor
+    /// terminates, `r`'s [`exit_hook`][Entity::exit_hook] will be called in the context of the
+    /// actor's root facet.
     pub fn add_exit_hook<M: 'static + Send>(&mut self, r: &Arc<Ref<M>>) {
         let r = Arc::clone(r);
         self.exit_hooks.push(Box::new(
